(* Ocsigen modules *)
open Eliom
open Eliomduce.Xhtml
open Lwt
open Simplexmlparser

(* Dose2 modules *)
open Napkin
open Rapids
open Diagnosis

open Anla_tools
open Packetosophy

let db = ref (create_database ());;
let (bundles: (string * string list * health) list ref) = ref [];;
let unit_index = get_unit_index !db;;
let version_index = get_version_index !db;;
let release_index = get_release_index !db;;
let architecture_index = get_architecture_index !db;;
let archive_index = get_archive_index !db;;
let source_index = get_source_index !db;;

let home_service =
	new_service ~url:[""] ~get_params:unit ();;
let list_archives_service =
	new_service ~url:["list_archives"] ~get_params:unit ();;
let show_archive_service =
	new_service ~url:["show_archive"] ~get_params:(string "name") ();;
let archive_chunk_service =
	new_service ~url:["archive"] ~get_params:(string "name" ** string "date" ** int "start") ();;
let archive_service =
	new_service ~url:["archive"] ~get_params:(string "name" ** string "date") ();;
let unit_info_service =
	new_service ~url:["unit_info"] ~get_params:(string "unit") ();;
let package_info_service =
	new_service ~url:["package_info"] ~get_params:(int "id") ();;
(* let source_info_service =
	new_service ~url:["source_info"] ~get_params:(string "source") ();; *)
let range_service =
	new_service ~url:["range"] ~get_params:(string "op" ** string "version" ** string "unit") ();; 
let lifetable_service = 
	new_service ~url:["lifetable"] ~get_params:(string "unit") ();;
let list_bundles_service =
	new_service ~url:["list_bundles"] ~get_params:unit ();;
let health_service =
	new_service ~url:["health"] ~get_params:(string "bundle") ();;

let anlaize ?(onload = "") sp (title: string) (body: Xhtml1_strict.flows): Xhtml1_strict.html =
{{ <html>[
		<head>[
			<title>{: title :}
			{{ css_link (make_uri (static_dir sp) sp ["default.css"]) }}
			{{ js_script (make_uri (static_dir sp) sp  ["functions.js"]) }}
		]
		<body onload={: onload :}>[
			<div class="head">[
				<h1>{: title :}
				<img style="position: absolute; top: 0; right: 0" alt="EDOS" src={: make_uri (static_dir sp) sp ["edos.png"] :}>[]
			]
			<div class="menu">
			[
				<table>[
					<tr>[<th>"Menu"]
					<tr>[<td>[{{ a home_service sp {{ "Home" }} () }}]]
					<tr>[<td>[{{ a (new_external_service ~url:["http://www.edos-project.org"] ~get_params:unit ~post_params:unit ()) sp {{ "EDOS Project" }} () }}] ]
				]
			]
			<div class="body">{: body :}
			<div class="foot">[
				<img src={: make_uri (static_dir sp) sp ["ocsigen.png"] :} alt="Ocsigen!">[]
			]
		]
	]	
}};;

let availability_list sp archives p_id =
	List.flatten (List.map (fun a_id ->
	let (intervals: Xhtml1_strict.flows ref) = ref {{ [] }} in
	let a_data = Archive_index.data archive_index a_id in
	let a_name = Archive_index.find archive_index a_id in	
		Lifetime.iterate_over_intervals (fun (j0, j1) ->
			let (y0, m0, d0) = Lifetime.ymd_of_day j0
			and (y1, m1, d1) = Lifetime.ymd_of_day j1 in
			let str0 = Format.sprintf "%04d-%02d-%02d" y0 m0 d0
			and str1 = Format.sprintf "%04d-%02d-%02d" y1 m1 d1 in
			intervals := {{
				{: !intervals :} @
				['from '	
				{: a archive_service sp {{ {: str0 :} }} (a_name, str0) :}
				' to '
				{: a archive_service sp {{ {: str1 :} }} (a_name, str1) :}] }}
		) (availability p_id a_data);
		if !intervals = {{ [] }} then
			[]
		else 
			[{{ <li>([
					'In '
					{: a show_archive_service sp {{ {: a_name :} }} a_name :}
					' '] @
					{: !intervals :}) 
				}}]
	) archives);;

let anla_welcome sp () () =
begin
	return (anlaize
		sp
		"Welcome to anla"
		{{[
				<div class="text">[
					<p style="color: red">"ANLA PROTOTYPE"
					<p>"The following functions are available:"
					<ul>[
						<li>[{: a list_bundles_service sp {{ "See day-by-day installability problems" }} () :}]
						<li>[{: a list_archives_service sp {{ "List known archives" }} () :}]
						<li>[{: a (new_external_service ~url:["http://ara.edos-project.org"] ~get_params:unit ~post_params:unit ()) sp {{ "Search packages using Ara" }} () :}]
						<li>[ {: get_form unit_info_service sp (fun unit_name ->
							{{ [<p>['Get information for the following unit: ' {: string_input ~input_type:{: "text" :} ~name:unit_name () :} ' ' {: string_input ~input_type:{: "submit" :} ~value:"Search" () :} ]] }}
						):} ]
					]
				]
				<div class="weather">[
					<table>[
						<tr>[<th>"Debian weather"]
					]
				]
		]}})
end;;

let anla_list_archives sp () () =
let generate_archive_row a_id =
let a_name = Archive_index.find archive_index a_id in	
begin
	[{{ [(a show_archive_service sp {: a_name :} (a_name))] }}]
end in 
begin
	Preemptive.detach Functions.archives !db >>=
	(fun archives -> Preemptive.detach (List.map generate_archive_row) (Archive_set.elements archives) >>=
	(fun archive_rows ->
	return (anlaize
		sp
		"Known archives"
		{{[
				<div class="text">(
					[<p>"Below you will find the archives that are known to the system."] @	
					[{: create_table [Red] ["Archive name"] archive_rows :}]
				)
			]
		}})))
end;;

let anla_show_archive sp name () =
let generate_date_rows a_data name date =
begin
	let (y, m, d) = Lifetime.ymd_of_day date in
	let date_string = Format.sprintf "%04d-%02d-%02d" y m d in 
	[ {{ [ {: (a archive_service sp {{ {: date_string :} }} (name, date_string)) :} ] }};
	  {{ {: string_of_int (Package_set.cardinal (get_archive_contents a_data date)) :} }}
	]
end in
begin
catch
	(fun () -> Preemptive.detach (Archive_index.search archive_index) name >>=
	(fun a_id -> Preemptive.detach (get_archive !db) a_id >>=
	(fun a_data -> Preemptive.detach (archive_range !db) a_data >>=
	(fun (d0, d1) -> let dates = ref [] in
		begin
			for i = d1 downto d0
			do
				dates := i::!dates
			done;
			Preemptive.detach (List.map (generate_date_rows a_data name)) !dates
		end	>>=
	(fun daterows ->
		return (anlaize
			sp
			("Packages in archive " ^ name)
			{{[
				<div class="text">[
					{: create_table [Red; Yellow] ["Date"; "Number of packages"] daterows :}
				] 
			]
		}}))))))
	(function
	  Not_found -> return (anlaize
		sp
		("Archive not found")
		{{[
			<div class="text">[
				<p>{: "The archive '" ^ name ^ "' is not known to the system." :}
			]
		]}}
	)
	| exn -> fail exn)
end;;

let anla_archive_chunk sp (name, (date, start)) () =
let sort_package_set pset =
let plist = Package_set.elements pset in
begin
	List.sort (fun (u1,_,_) (u2,_,_) -> String.compare u1 u2)
	(List.map (fun p_id ->
		let pkg = Functions.get_package_from_id !db p_id in
		let unit_name = Unit_index.find unit_index pkg.pk_unit in
		let v = Version_index.get_version (fst pkg.pk_version)
		and r = Release_index.get_version (snd pkg.pk_version) in
			(unit_name, v, r)) plist)
end in
begin
	let chunk_size = 20 in
	catch
	(fun () -> Preemptive.detach (Archive_index.search archive_index) name >>=
	(fun a_id -> Preemptive.detach (get_archive !db) a_id >>=
	(fun a_data -> Preemptive.detach (get_archive_contents a_data) (Scanf.sscanf date "%04d-%02d-%02d" (fun y m d -> Lifetime.day_of_ymd (y,m,d))) >>=
	(fun p_set -> Preemptive.detach sort_package_set p_set >>=
	(fun p_list -> 
		return (anlaize
			sp
			(Format.sprintf "Packages in archive %s on %s" name date)
			{{[
					<div class="text">[
						<p>(
							{: if start < chunk_size then {{ "Previous" }} else {{ [ {: a archive_chunk_service sp {{ "[Previous]" }} (name, (date, start - chunk_size)) :} ] }} :} @
							" " @
							{: if start > List.length p_list - chunk_size then {{ "Next" }} else {{ [ {: a archive_chunk_service sp {{ "[Next]" }} (name, (date, start + chunk_size)) :} ] }} :}
						)
						{: create_table [Red; Yellow] ["Unit"; "Version"] (List.map (fun (unit_name, v, r) ->
							[ {{ [ {: a unit_info_service sp {{ {: unit_name :} }} unit_name :} ] }};
								{{ {: scribe_version v r :} }}]) (list_cut p_list start chunk_size)) :} 
					]
				]
			}}))))))
	(function
	  Not_found -> return (anlaize
		sp
		("Archive not found")
		{{[
			<div class="text">[
				<p>{: "The archive '" ^ name ^ "' is not known to the system." :}
			]
		]}})
	| exn -> fail exn)
end;;

let anla_archive sp (name, date) () =
	anla_archive_chunk sp (name, (date, 0)) ();;

let anla_unit_info sp name () =
let create_unit_row archives (p_id, pkg) =
let version = Version_index.get_version (fst pkg.pk_version)
and release = Release_index.get_version (snd pkg.pk_version) in
let arch = Architecture_index.find architecture_index pkg.pk_architecture in
	[{{ {: name :} }};
	 {{ {: scribe_version version release :} }};
   {{ {: arch :} }};
   {{ [ {: a package_info_service sp {{ "link" }} (Package_ID.to_int p_id) :} ] }};
   match (availability_list sp archives p_id) with
	 	[] -> {{ "Not available" }}
	| h::t -> {{ [<ul>([{: h :}] @ {: t :})] }}
		] in
begin
	catch
	(fun () -> Preemptive.detach Archive_set.elements (Functions.archives !db) >>=
	(fun archives -> Preemptive.detach (Unit_index.search unit_index) name >>=
	(fun u_id -> Preemptive.detach (Functions.unit_id_to_package_set !db) u_id >>=
	(fun pkgs -> Preemptive.detach (List.map (create_unit_row archives)) (List.sort (cmp_pkgs version_index release_index) (List.map (fun x -> (x, Functions.get_package_from_id !db x)) (Package_set.elements pkgs))) >>= 
	(fun pkgrows -> Preemptive.detach (Functions.unit_id_to_providers !db) u_id >>=
	(fun providers -> Preemptive.detach (List.filter (fun (p, _) -> not (Package_set.mem p pkgs))) providers >>=
	(fun real_providers -> Preemptive.detach (List.map (fun (p, _) -> 
		let pkg = Functions.get_package_from_id !db p in
			(Package_ID.to_int p, Unit_index.find unit_index pkg.pk_unit, scribe_version (Version_index.get_version (fst pkg.pk_version)) (Release_index.get_version (snd pkg.pk_version))))) real_providers >>=  
	(fun provrows ->
		return (anlaize
			sp
			("Information for unit " ^ name)
			{{[
				<div class="text">([
					<h2>"Lifetable"
					<p>[{: a lifetable_service sp {{ "See the lifetable" }} name :}]
					<h2>"Versions and packages"] @	
					{: if List.length pkgrows = 0 then
							{{ [<p>{: Format.sprintf "Unit \"%s\" does not exist as a package." name :}] }} else
							{{ [ {: create_table [Red;Yellow;Green;Cyan;Blue] ["Unit";"Version";"Architecture";"Package";"Availability"] pkgrows :} ] }} :} @
					{: if List.length provrows = 0 then 
							{{ [<p>{: Format.sprintf "No other package provides the unit \"%s\"" name :}] }}
							else
							{{ [<h2>{: Format.sprintf "Other packages providing the unit \"%s\"" name :}
								<ul>([<li>[{: let (p_id, n, v) = List.hd provrows in
									a package_info_service sp {{ {: Format.sprintf "%s %s" n v :} }} p_id :}]] @
								{: List.map (fun (p_id, n, v) -> {{ <li>[{: a package_info_service sp {{ {: Format.sprintf "%s %s" n v :} }} p_id :}] }}) (List.tl provrows) :})
								] }} :}
				)
			]}})))))))))
	(function
		Not_found -> return (anlaize
			sp
			"Unit not found"
		{{[
			<div class="text">[
				<p>{: "The unit '" ^ name ^ "' is not known to the system." :}
			]
		]}})
	| exn -> fail exn)
end;;

let anla_package_info sp id () =
let version_of (v, r): string =
let v_str = Version_index.get_version v
and r_str = Release_index.get_version r in
	scribe_version v_str r_str in
let create_versioned (add_any_of: bool) v_list =
begin
	Preemptive.detach (List.map (fun (u, sel) ->
		let unit_name = Unit_index.find unit_index u in
		{{ [{: a unit_info_service sp {{ {: unit_name :} }} unit_name :} 
				' ('
			  {: match sel with
					Sel_LEQ v -> {{ {: a range_service sp {{ {: (Format.sprintf "<= %s" (version_of v)) :} }} ("leq", (version_of v, unit_name)) :} }}
				| Sel_GEQ v -> {{ {: a range_service sp {{ {: (Format.sprintf ">= %s" (version_of v)) :} }} ("geq", (version_of v, unit_name)) :} }}
				| Sel_EQ v -> {{ {: a range_service sp {{ {: (Format.sprintf "= %s" (version_of v)) :} }} ("eq", (version_of v, unit_name)) :} }}
				| Sel_LT v -> {{ {: a range_service sp {{ {: (Format.sprintf "<< %s" (version_of v)) :} }} ("lt", (version_of v, unit_name)) :} }}
				| Sel_GT v -> {{ {: a range_service sp {{ {: (Format.sprintf ">> %s" (version_of v)) :} }} ("gt", (version_of v, unit_name)) :} }}
				| Sel_ANY -> {{ <i>"any version" }} :}
				')']
		}}
		)) v_list >>=
	(fun units -> return
		(if List.length units = 0 then {{ "" }}
		else if List.length units = 1 then {{ {: List.hd units :} }}
		else
			{{ ({: if add_any_of then {{ ['Any of:'] }} else {{ [] }} :} @ [<ul>([<li>{: List.hd units :}] @ {: List.map (fun x -> {{ <li>{: x :} }}) (List.tl units) :})]) }}))
end in
let create_versioned_disj d_list =
begin
	lwt_map (create_versioned true) d_list >>=
	(fun deps -> return
		(if List.length deps = 0 then {{ "" }}
		else if List.length deps = 1 then {{ {: List.hd deps :} }}
		else
			{{ ['All of:' <ul>([<li>{: List.hd deps :}] @ {: List.map (fun x -> {{ <li>{: x :} }}) (List.tl deps) :})] }} 
	))
end in
begin
	catch
	(fun () ->
		Preemptive.detach Package_ID.from_int id >>=
		(fun p_id -> Preemptive.detach (Functions.get_package_from_id !db) p_id >>=
		(fun pkg -> Preemptive.detach (Unit_index.find unit_index) pkg.pk_unit >>=
		(fun unit_name -> Preemptive.detach Version_index.get_version (fst pkg.pk_version) >>=
		(fun version -> Preemptive.detach Release_index.get_version (snd pkg.pk_version) >>=
		(fun release -> Preemptive.detach Archive_set.elements (Functions.archives !db) >>=
		(fun archives -> Preemptive.detach (Source_index.find source_index) pkg.pk_source >>=
		(fun (source_name, source_version) -> create_versioned false pkg.pk_provides >>=
		(fun provides -> create_versioned false pkg.pk_conflicts >>=
		(fun conflicts -> create_versioned false pkg.pk_replaces >>=
		(fun replaces -> create_versioned_disj pkg.pk_depends >>=
		(fun depends -> create_versioned_disj pkg.pk_pre_depends >>=
		(fun pre_depends -> Preemptive.detach (availability_list sp archives) p_id >>=
		(fun availability ->
		return (anlaize
			sp
			"Package information"
			{{[
								{: create_table [Red; Yellow] ["Field"; "Contents"]
									[[{{ "Package" }} ; {{ {: string_of_int id :} }}];
									 [{{ "Unit" }}; {{ [ {: a unit_info_service sp {{ {: unit_name :} }} unit_name :} ] }}];
									 [{{ "Source" }}; {{ {: source_name :} }}];
									 [{{ "Version" }}; {{ {: scribe_version version release :} }}];
									 [{{ "Essential" }}; {{ {: if pkg.pk_essential then "yes" else "no" :} }} ];
									 [{{ "Build-Essential" }}; {{ {: if pkg.pk_build_essential then "yes" else "no" :} }} ];
									 [{{ "Provides" }}; provides];
									 [{{ "Conflicts" }}; conflicts];
									 [{{ "Replaces" }}; replaces];
									 [{{ "Depends" }}; depends];
									 [{{ "Pre-Depends" }}; pre_depends];
									 [{{ "Availability" }}; match availability with
												[] -> {{ "Not available" }}
											| h::t -> {{ [<ul>([{: h :}] @ {: t :})] }}
									 ]] :}]}} )))))))))))))))
					(function 
						Not_found -> return (anlaize
							sp
							"Package not found"
						{{[
							<div class="text">[
								<p>{: "The package is not known to the system." :}
							]
						]}})
					| exn -> fail exn)
				end;;

				(* let anla_source_info sp name () =
				begin
					catch 
					(fun () -> Preemptive.detach get_source_index !db >>=
					(fun source_index -> Preemptive.detach (Source_index.search source_index) name >>=
					return (anlaize
						sp
						(Format.sprintf "Information for source unit \"%s\"" name)
						{{ {: create_table [Red; Yellow] ["Version"; "Generated packages"] :} }}
	)))
	(function 
		Not_found -> return (anlaize
			sp
			"Source unit not found"
		{{[
			<div class="text">[
				<p>{: Format.sprintf "The source unit \"%s\" is not known to the system." name :}
			]
		]}})
	| exn -> fail exn) 
end;; *)

let anla_range sp (op, (version, unit_name)) () =
let selector v =
	if op = "any" then Sel_ANY
	else if op = "lt" then Sel_LT v
	else if op = "leq" then Sel_LEQ v
	else if op = "eq" then Sel_EQ v
	else if op = "geq" then Sel_GEQ v
	else if op = "gt" then Sel_GT v
	else raise (Failure ("Unknown operator " ^ op)) in
let op_string_of op =
	if op = "lt" then "<<"
	else if op = "leq" then "<="
	else if op = "eq" then "="
	else if op = "geq" then ">="
	else if op = "gt" then ">>"
	else op in
let create_pkg_list range =
begin
	let x = List.map (fun (p_id, pkg) ->
		let unit_name = Unit_index.find unit_index pkg.pk_unit in
		let v = Version_index.get_version (fst pkg.pk_version)
		and r = Release_index.get_version (snd pkg.pk_version) in
		{{ <li>[ {: a unit_info_service sp {{ {: unit_name :} }} unit_name :} 
			' '
			{: a package_info_service sp {{ {: scribe_version v r :} }} p_id :} ] }}
		) (List.sort (cmp_pkgs version_index release_index) (List.map (fun p_id -> (Package_ID.to_int p_id, Functions.get_package_from_id !db p_id)) (Package_set.elements range))) in
	match x with
		[] -> {{ [<p>"Unmatched"] }}
	| h::t -> {{ [<p>"Matched by" <ul>([{: h :}] @ {: t :})] }}
end in
begin
	catch
	(fun () -> Preemptive.detach (Unit_index.search unit_index) unit_name >>=
	(fun u_id -> Preemptive.detach Rapids.split_version version >>=
	(fun (v, r) -> Preemptive.detach (Version_index.search version_index) v >>=
	(fun v_id -> Preemptive.detach (Release_index.search release_index) r >>=
	(fun r_id -> Preemptive.detach (Functions.select !db) (u_id, (selector (v_id, r_id))) >>=
	(fun range -> Preemptive.detach create_pkg_list range >>=
	(fun pkg_list ->
		return (anlaize 
			sp
			"Range"
			{{ ([<p>(['Packages matching the range ' 
				{: a unit_info_service sp {{ {: unit_name :} }} unit_name :}] @
				{: Format.sprintf " (%s %s)" (op_string_of op) version :})] @
				{: pkg_list :}
				) }}
	))))))))
	(function
		exn -> fail exn)
end;;

let anla_lifetable sp unit_name () =
let put_version p_id =
	let pkg = Functions.get_package_from_id !db p_id in
	let v = Version_index.get_version (fst pkg.pk_version)
	and r = Release_index.get_version (snd pkg.pk_version) in
		scribe_version v r in
let rec really_range_over_versions light v_list =
begin
	match v_list with
		[] -> []
	| h::t -> {{ <tr>[<th style={: Format.sprintf "background-color: %s" (if light then "ddddff" else "d0d0ff") :}>{: put_version h :}] }}::(really_range_over_versions (not light) t)
end in
let range_over_versions row_colour a_name v_list =
begin
	match v_list with
		[] -> []
	| h::t -> {{ <tr>[<th style={: Format.sprintf "background-color: %s" row_colour :} rowspan={: string_of_int (List.length v_list) :}>{: a_name :} <th style="background-color: ddddff">{: put_version h :}] }}::(really_range_over_versions false t)
end in
let rec range_over_archives light v_list arch_list =
begin
	match arch_list with
		[] -> []
	| h::t -> let a_name = Archive_index.find archive_index h in
			(range_over_versions (if light then "8a8aff" else "9f9fff") a_name v_list)::(range_over_archives (not light) v_list t)
end in
let create_lifetable_headers v_list arch_list =
begin
	{{ [<table id="lifetable">([<tr>[<th>"Archive" <th>"Version"]] @
	{: List.flatten (range_over_archives true v_list arch_list) :})]	}}
end in
let array_of_ranges archives v_list =
begin
	Format.sprintf "new Array(%s)" (String.concat "," (List.flatten (List.map (fun a_id ->
		List.map (fun p_id ->
			let str = ref [] in
			let av = availability p_id (Rapids.get_archive !db a_id) in
			Lifetime.iterate_over_intervals (fun (s, e) ->
				str := (Format.sprintf "new range(%d,%d)" s e)::!str) av;
				Format.sprintf "new Array(%s)" (String.concat "," (List.rev !str))
		) v_list
	) archives)))
end in
begin
	catch
	(fun () -> Preemptive.detach Archive_set.elements (Functions.archives !db) >>=
	fun archives -> Preemptive.detach (Unit_index.search unit_index) unit_name >>=
	fun u_id -> Preemptive.detach (Functions.unit_id_to_package_set !db) u_id >>=
	fun prov_set -> Preemptive.detach Package_set.elements prov_set >>=
	fun prov_list -> Preemptive.detach (create_lifetable_headers prov_list) archives >>=
	fun archive_table -> return (anlaize 
		~onload:(Format.sprintf "lifetable(%s)" (array_of_ranges archives prov_list))
		sp
		(Format.sprintf "Lifetable for unit %s" unit_name)
		{: archive_table :}
	))
	(function
		exn -> fail exn)
end;;

let anla_list_bundles sp () () =
begin
	catch
	(fun () -> Preemptive.detach (fun x -> 
		List.map (fun (name, archives, health) ->
			[{{ [{: a health_service sp {{ {:  name :} }} name :}] }}; {{ {: String.concat " + " archives :} }}] 
		) x) !bundles >>=
	fun bundle_rows -> return (anlaize
		sp
		"Available bundles"
		{{ [ {: create_table [Red;Yellow] ["Name";"Contents"] bundle_rows :} ] }}
	))
	(function
		exn -> fail exn)
end;;

let anla_health sp bundle_name () =
let rec find_bundle b bundle_name =
begin
	match b with
		[] -> raise Not_found
	| (n,c,h)::xs -> if n = bundle_name then (c,h) else find_bundle xs bundle_name
end in
begin
	catch
	(fun () -> Preemptive.detach (find_bundle !bundles) bundle_name >>=
	fun (contents, h_tbl) -> Preemptive.detach (fun (h: Packetosophy.health) ->
		let res = ref [] in
		let reasons = Hashtbl.find h (Lifetime.day_of_ymd (2007, 6, 30)) in
		Hashtbl.iter (fun p_id rns ->
			let pkg = Functions.get_package_from_id !db p_id in
			let name = Unit_index.find unit_index pkg.pk_unit in
			let v = Version_index.get_version (fst pkg.pk_version) in
			let r = Release_index.get_version (snd pkg.pk_version) in
			let lis = List.map (fun rsn ->
				match rsn with
				| Not_available p_id' ->
					let pkg' = Functions.get_package_from_id !db p_id' in
					let name' = Unit_index.find unit_index pkg'.pk_unit in
					let v' = Version_index.get_version (fst pkg'.pk_version) in
					let r' = Release_index.get_version (snd pkg'.pk_version) in
					{{ <li>{: Format.sprintf "package %s-%s not available" name' (scribe_version v' r') :} }}
				| Requested p_id' ->
					let pkg' = Functions.get_package_from_id !db p_id' in
					let name' = Unit_index.find unit_index pkg'.pk_unit in
					let v' = Version_index.get_version (fst pkg'.pk_version) in
					let r' = Release_index.get_version (snd pkg'.pk_version) in
					{{ <li>{: Format.sprintf "package %s-%s requested" name' (scribe_version v' r') :} }}
				| Dependency (p_id', p_ids) ->
					let dep_str = String.concat ", " (List.map (fun p_id'' ->
						let pkg'' = Functions.get_package_from_id !db p_id'' in
						let name'' = Unit_index.find unit_index pkg''.pk_unit in
						let v'' = Version_index.get_version (fst pkg''.pk_version) in
						let r'' = Release_index.get_version (snd pkg''.pk_version) in
						Format.sprintf "%s-%s" name'' (scribe_version v'' r'')) p_ids) in
					{{ <li>{: Format.sprintf "depends on %s" dep_str :} }}
				| Empty_disjunction (p_id', vs) ->
					let pkg' = Functions.get_package_from_id !db p_id' in
					let name' = Unit_index.find unit_index pkg'.pk_unit in
					let v' = Version_index.get_version (fst pkg'.pk_version) in
					let r' = Release_index.get_version (snd pkg'.pk_version) in
					let v_str = String.concat ", " (List.map (fun (u_id'', s) ->
						let name'' = Unit_index.find unit_index u_id'' in
						match s with
						| Sel_ANY -> name''
						| Sel_LT (v'', r'') -> Format.sprintf "%s (< %s)" name'' (scribe_version (Version_index.get_version v'') (Release_index.get_version r''))
						| Sel_LEQ (v'', r'') -> Format.sprintf "%s (<= %s)" name'' (scribe_version (Version_index.get_version v'') (Release_index.get_version r''))
						| Sel_EQ (v'', r'') -> Format.sprintf "%s (= %s)" name'' (scribe_version (Version_index.get_version v'') (Release_index.get_version r''))
						| Sel_GEQ (v'', r'') -> Format.sprintf "%s (>= %s)" name'' (scribe_version (Version_index.get_version v'') (Release_index.get_version r''))
						| Sel_GT (v'', r'') -> Format.sprintf "%s (> %s)" name'' (scribe_version (Version_index.get_version v'') (Release_index.get_version r''))
					) vs) in
					{{ <li>{: Format.sprintf "missing dependency: (%s-%s) %s" name' (scribe_version v' r') v_str :} }}
				| Conflict (p_id1, p_id2) ->	
					let pkg1 = Functions.get_package_from_id !db p_id1 in
					let name1 = Unit_index.find unit_index pkg1.pk_unit in
					let v1 = Version_index.get_version (fst pkg1.pk_version) in
					let r1 = Release_index.get_version (snd pkg1.pk_version) in
					let pkg2 = Functions.get_package_from_id !db p_id2 in
					let name2 = Unit_index.find unit_index pkg2.pk_unit in
					let v2 = Version_index.get_version (fst pkg2.pk_version) in
					let r2 = Release_index.get_version (snd pkg2.pk_version) in
					{{ <li>{: Format.sprintf "packages %s-%s and %s-%s conflict" name1 (scribe_version v1 r1) name2 (scribe_version v2 r2) :} }}
			) rns in
			res := (Format.sprintf "%s-%s is not installable:" name (scribe_version v r), {{ <ul>[{: List.hd lis :} !{: List.tl lis :}] }})::!res
		) reasons;
		List.map (fun (n, u) -> {{ <li>[!{: n :} {: u :}] }}) !res
	) h_tbl >>=
	fun res -> return (anlaize
		sp
		(Format.sprintf "Health status of bundle %s" bundle_name)
		{{ [<ul>[{: List.hd res :} !{: List.tl res :}]] }} 
	))
	(function
		exn -> fail exn)
end;;

let read_waterways () =
begin
	List.iter (fun cfg ->
		match cfg with
		| Element ("source", [("spec", s)], _) ->
				let spec = Waterway.specification_of_string s in
				begin
					Messages.warning ("Merging waterway " ^ s);
					Waterway.merge !db spec
				end
		| Element ("bundles", _, bundles_spec) ->
				List.iter (fun b ->
					match b with
					|	Element ("bundle", [("name", bundle_name)], archives_spec) ->
						let bundle_contents = ref [] in
						begin
							List.iter (fun a ->
								match a with
								| Element ("archive", [("name", archive_name)], _) ->
									bundle_contents := archive_name::!bundle_contents;
								| _ -> raise (Extensions.Error_in_config_file "Unexpected content inside Dose config (bundle: archive expected)")
							) archives_spec;
							bundles := (bundle_name, !bundle_contents, Hashtbl.create 1)::!bundles
						end
					| _ -> raise (Extensions.Error_in_config_file "Unexpected content inside Dose configuration (bundles: bundle expected)")
				) bundles_spec
		| _ -> raise (Extensions.Error_in_config_file "Unexpected content inside Dose configuration (source or bundles expected)")
	) (Eliom.get_config ());
	List.iter (fun (name, bundle, health) ->
		Messages.debug (Format.sprintf "Examining health for bundle %s" name);
		examine_health !db bundle health
	) !bundles
end;;

let _  = 
begin
	read_waterways ();
	register ~service:home_service anla_welcome;
	register ~service:list_archives_service anla_list_archives;
	register ~service:show_archive_service anla_show_archive;
	register ~service:archive_chunk_service anla_archive_chunk;
	register ~service:archive_service anla_archive;
	register ~service:unit_info_service anla_unit_info;
	register ~service:package_info_service anla_package_info;
	(* register ~service:source_info_service anla_source_info *)
	register ~service:range_service anla_range;
	register ~service:lifetable_service anla_lifetable;
	register ~service:list_bundles_service anla_list_bundles;
	register ~service:health_service anla_health
end;;
