(* Fragments *)

exception Error of string;;

let sf = Printf.sprintf;;

(*** tag *)
type tag =
  | T_More
  | T_End
  | T_Data
  | T_Stop_data
  | T_Ident
  | T_Array
  | T_Record
  | T_Sum
  | T_List
  | T_Hashtbl
  | T_Collection
;;
(* ***)
(*** tags *)
let tags = [
  T_More;
  T_End;
  T_Data;
  T_Stop_data;
  T_Ident;
  T_Array;
  T_Sum;
  T_Record;
  T_List;
  T_Hashtbl;
  T_Collection;
];;
(* ***)
(*** string_of_tag *)
let string_of_tag = function
| T_More       -> "more"
| T_End        -> "end"
| T_Data       -> "data"
| T_Stop_data  -> "stopdata"
| T_Ident      -> "ident"
| T_Array      -> "array"
| T_Record     -> "record"
| T_Sum        -> "sum"
| T_List       -> "list"
| T_Hashtbl    -> "hashtbl"
| T_Collection -> "collection"
;;
(* ***)
(*** short_string_of_tag *)
let short_string_of_tag = function
| T_More       -> "m"
| T_End        -> "e"
| T_Data       -> "d"
| T_Stop_data  -> "s"
| T_Ident      -> "i"
| T_Array      -> "a"
| T_Record     -> "r"
| T_Sum        -> "s"
| T_List       -> "l"
| T_Hashtbl    -> "h"
| T_Collection -> "c"
;;
(* ***)
(*** tag_of_string *)
let tag_of_string = function
| "more"       -> T_More
| "end"        -> T_End
| "data"       -> T_Data
| "stopdata"   -> T_Stop_data
| "ident"      -> T_Ident
| "array"      -> T_Array
| "record"     -> T_Record
| "sum"        -> T_Sum
| "list"       -> T_List
| "hashtbl"    -> T_Hashtbl
| "collection" -> T_Collection
| x -> raise (Error(sf "Unknown tag %S" x))
;;
(* ***)
(*** int_of_tag *)
let int_of_tag = function
| T_More       -> 0
| T_End        -> 1
| T_Data       -> 2
| T_Stop_data  -> 3
| T_Ident      -> 4
| T_Array      -> 5
| T_Record     -> 6
| T_Sum        -> 7
| T_List       -> 8
| T_Hashtbl    -> 9
| T_Collection -> 10
;;
(* ***)
(*** tag_of_int *)
let tag_of_int = function
| 0 -> T_More
| 1 -> T_End
| 2 -> T_Data
| 3 -> T_Stop_data
| 4 -> T_Ident
| 5 -> T_Array
| 6 -> T_Record
| 7 -> T_Sum
| 8 -> T_List
| 9 -> T_Hashtbl
| 10 -> T_Collection
| x -> raise (Error(sf "Unknown tag %d" x))
;;
(* ***)
(*** is_start_tag *)
let is_start_tag = function
| T_Array|T_Data|T_List|T_Hashtbl|T_Record|T_Collection -> true
| _ -> false
;;
(* ***)
(*** is_end_tag *)
let is_end_tag = function
| T_End|T_Stop_data -> true
| _ -> false
;;
(* ***)
let max_tag = 32;;

type ident = string;;
(*** token *)
type token =
| Bool of bool
| Char of char
| Int of int
| Int64 of int64
| Float of float
| String of string
| Tag of tag
| Ident of ident
;;
(* ***)
(*** string_of_token *)
let string_of_token = function
| Bool b   -> sf "Bool(%b)" b
| Char c   -> sf "Char(%C)" c
| Int d    -> sf "Int(%d)" d
| Int64 d  -> sf "Int(%Ld)" d
| Float f  -> sf "Float(%f)" f
| String s -> sf "String(%S)" s
| Tag x    -> sf "Tag(%S)" (string_of_tag x)
| Ident x  -> sf "Ident(%S)" x
;;
(* ***)

(*** io_in *)
type io_in = {
  mutable io_lost : bool;
  io_read_token : (unit -> token);
  mutable io_finish : (unit -> unit)
}
(* ***)
(*** io_out *)
type io_out = {
  io_write_token : (token -> unit);
  io_flush : (unit -> unit)
};;
(* ***)

(*** read_token *)
let read_token i = i.io_read_token ();;
(* ***)
(*** read_tag *)
let read_tag i =
  match read_token i with
  | Tag tag -> tag
  | x -> raise (Error(sf "Expecting tag got %s" (string_of_token x)))
;;
(* ***)
(*** read_this_tag *)
let read_this_tag i tag =
  let tag' = read_tag i in
  if tag = tag' then
    ()
  else
    raise (Error(sf "Expecting tag <%s> got <%s>"
                    (string_of_tag tag)
                    (string_of_tag tag')))
;;
(* ***)
(*** start *)
let start i = read_this_tag i T_Data;;
(* ***)
(*** default_finish *)
let default_finish i () = read_this_tag i T_Stop_data;;
(* ***)
(*** create_io_in *)
let create_io_in ~read_token ?finish () =
  let i =
    { io_lost = false;
      io_read_token = read_token;
      io_finish = ignore }
  in
  i.io_finish <-
    begin
      match finish with
      | None -> default_finish i
      | Some f -> f i
    end;
  start i;
  i
;;
(* ***)
(*** finish *)
let finish i = i.io_finish ();;
(* ***)
(*** read_int *)
let read_int i =
  match read_token i with
  | Int x -> x
  | x -> raise (Error(sf "Expecting int got %s" (string_of_token x)))
;;
(* ***)
(*** read_int64 *)
let read_int64 i =
  match read_token i with
  | Int64 x -> x
  | x -> raise (Error(sf "Expecting int64 got %s" (string_of_token x)))
;;
(* ***)
(*** read_float *)
let read_float i =
  match read_token i with
  | Float x -> x
  | x -> raise (Error(sf "Expecting float got %s" (string_of_token x)))
;;
(* ***)
(*** read_bool *)
let read_bool i =
  match read_token i with
  | Bool x -> x
  | x -> raise (Error(sf "Expecting bool got %s" (string_of_token x)))
;;
(* ***)
(*** read_char *)
let read_char i =
  match read_token i with
  | Char x -> x
  | x -> raise (Error(sf "Expecting char got %s" (string_of_token x)))
;;
(* ***)
(*** read_ident *)
let read_ident i =
  match read_token i with
  | Ident x -> x
  | x -> raise (Error(sf "Expecting ident got %s" (string_of_token x)))
;;
(* ***)
(*** read_string *)
let read_string i =
  match read_token i with
  | String x -> x
  | x -> raise (Error(sf "Expecting string got %s" (string_of_token x)))
;;
(* ***)

(*** write_{token,int,char,string,bool,float,tag,ident} *)
let write_token o t = o.io_write_token t;;
let write_int o x    = write_token o (Int x)
let write_int64 o x  = write_token o (Int64 x)
let write_char o x   = write_token o (Char x)
let write_string o x = write_token o (String x)
let write_bool o x   = write_token o (Bool x)
let write_float o x  = write_token o (Float x)
let write_tag o x    = write_token o (Tag x)
let write_ident o x  = write_token o (Ident x)
(* ***)
(*** loss *)
let loss i = i.io_lost <- true;;
(* ***)
(*** lost *)
let lost i = i.io_lost;;
(* ***)
(*** create_io_out *)
let create_io_out ~write_token ?(flush=ignore) () =
  let o =
    { io_write_token = write_token;
      io_flush = flush }
  in
  write_tag o T_Data;
  o
;;
(* ***)
(*** default_flush *)
let default_flush o = write_tag o T_Stop_data;;
(* ***)
(*** flush *)
let flush o =
  default_flush o;
  o.io_flush ()
;;
(* ***)
