(* Human *)

open Io;;
open Fragments;;

let sf = Printf.sprintf;;

(*** G *)
module G =
  struct
    include Genlex
    (*** string_of_token *)
    let string_of_token = function
    | Ident x  -> sf "Ident(%S)" x
    | Int d    -> sf "Int(%d)" d
    | Float f  -> sf "Float(%f)" f
    | Kwd x    -> sf "Kwd(%S)" x
    | String s -> sf "String(%S)" s
    | Char c   -> sf "Char(%C)" c
    (* ***)
  end
;;
(* ***)
(*** tag_list *)
let tag_list =
  ["true";"false"] @
  (List.map string_of_tag tags)
;;
let short_tag_list =
  ["t";"f"] @
  (List.map short_string_of_tag tags)
;;
(* ***)
(*** in_of_stream *)
let in_of_stream ?(short=true) ~stream () =
  let g = G.make_lexer (if short then short_tag_list else tag_list) stream in
  let next () =
    try
      Stream.next g
    with
    | Stream.Failure ->
        Stream.empty g;
        raise End_of_file
  in
  (*let peek () = Stream.peek g in
  let junk () = Stream.junk g in*)
  create_io_in
    ~read_token:
      begin fun () ->
        match next () with
        | G.Ident x -> 
          let m = String.length x in
          if m = 0 or x.[0] <> '_' then
            raise (Error(sf "Invalid ident %S" x))
          else
            Ident(String.sub x 1 (m - 1))
        | G.Int x       -> Int x
        | G.Char x      -> Char x
        | G.Float x     -> Float x
        | G.String x    -> String x
        | G.Kwd "true"  -> Bool true
        | G.Kwd "false" -> Bool false
        | G.Kwd "t"  -> Bool true
        | G.Kwd "f" -> Bool false
        | G.Kwd x       -> Tag(tag_of_string x)
      end
      ()
;;
(* ***)
(*** in_of_channel *)
let in_of_channel ?short ~in_channel () =
  in_of_stream ?short ~stream:(Stream.of_channel in_channel) ()
;;
(* ***)
(*** out_of_formatter *)
let out_of_formatter ?(short=true) ~formatter ?(indent=false) () =
  let fp = Format.fprintf in
  let fmt = formatter in
  let space = ref false in
  let string_of_tag = if short then short_string_of_tag else string_of_tag in
  create_io_out
    ~write_token:
      begin function
      | Int x -> fp fmt "@ %d" x
      | Int64 x -> fp fmt "@ %Ld" x
      | Char x -> fp fmt "@ %C" x
      | String x -> fp fmt "@ %S" x
      | Float x -> fp fmt "@ %F" x
      | Bool x -> fp fmt "@ %b" x
      | Tag x ->
          if indent && is_end_tag x then fp fmt "@]";
          if !space then
            fp fmt "@ %s" (string_of_tag x)
          else
            fp fmt "%s" (string_of_tag x);
          space := true;
          if indent && is_start_tag x then fp fmt "@["
      | Ident x -> fp fmt "@ _%s" x
      end
    ~flush:
      begin fun () ->
        fp fmt "@."
      end
    ()
;;
(* ***)
(*** out_of_channel *)
let out_of_channel ?short ~out_channel =
  out_of_formatter ?short ~formatter:(Format.formatter_of_out_channel out_channel)
;;
(* ***)
