(* Io *)

open Fragments;;

exception Unknown_tag;;

type io_in = Fragments.io_in
type io_out = Fragments.io_out

(*** r, w, literate, sum_io_spec, record_io_spec *)
type 'a r = io_in -> 'a
type 'a w = 'a -> io_out -> unit
type 'a literate = 'a r * 'a w
type 'a sum_io_spec = (string -> 'a r) * ('a -> string * (io_out -> unit));;
type 'a record_io_spec = (string * ('a -> io_in -> 'a) * ('a -> io_out -> unit)) list;;
type ('thing,'bag) iterator = 'bag -> ('thing -> unit) -> unit
type ('thing,'bag) builder = unit -> ('thing -> unit) * (unit -> 'bag)
type ('thing,'bag) collection_io_spec = 'thing literate * ('thing,'bag) iterator * ('thing,'bag) builder
(* ***)
(*** read, write, flush *)
let read : 'a literate -> io_in -> 'a = fun (r,_) i -> r i;;
let write : 'a literate -> 'a -> io_out -> unit = fun (_,w) x o -> w x o;;
let flush : io_out -> unit = fun o -> Fragments.flush o;;
let finish i = finish i;;
let write_and_flush io x o = write io x o; flush o;;
(* ***)
(*** convert *)
let convert ~(reader : io_in) ~(writer : io_out) () =
  let rec loop () =
    let t = read_token reader in
    if t <> Tag T_Stop_data then
      begin
        write_token writer t;
        loop ()
      end
    else
      flush writer
  in
  loop ()
;;
(* ***)
(*** dump *)
let dump ~(reader : io_in) () =
  try
    let i = ref 0 in
    while true do
      let t = read_token reader in
      Printf.printf "%d %s\n%!" !i (string_of_token t);
      incr i
    done
  with
  | End_of_file -> finish reader
;;
(* ***)
(*** io_{int,string,char,bool,float,unit,int64} *)
let io_int : int literate = (fun i -> read_int i), (fun x o -> write_int o x);;
let io_string : string literate = (fun i -> read_string i), (fun x o -> write_string o x);;
let io_char : char literate = (fun i -> read_char i), (fun x o -> write_char o x);;
let io_bool : bool literate = (fun i -> read_bool i), (fun x o -> write_bool o x);;
let io_float : float literate = (fun i -> read_float i), (fun x o -> write_float o x);;
let io_int64 : int64 literate = (fun i -> read_int64 i), (fun x o -> write_int64 o x);;
let io_unit : unit literate = (fun i -> ()), (fun x o -> ());;
(* ***)
(*** assoc3 *)
let rec assoc3 key = function
  | [] -> raise Not_found
  | (x,y,z)::rest when x = key -> (y,z)
  | _::rest -> assoc3 key rest
;;
(* ***)
(*** io_pair *)
let io_pair : 'a literate -> 'b literate -> ('a * 'b) literate = fun ((xr,xw) as xio) ((yr,yw) as yio) ->
  begin fun i ->
    let x = read xio i in
    let y = read yio i in
    (x, y)
  end,
  begin fun (x,y) o ->
    write xio x o;
    write yio y o
  end
;;
(* ***)
(*** io_triple *)
let io_triple : 'a literate -> 'b literate -> 'c literate -> ('a * 'b * 'c) literate =
  fun
    ((xr,xw) as xio)
    ((yr,yw) as yio)
    ((zr,zw) as zio)
  ->
  begin fun i ->
    let x = read xio i in
    let y = read yio i in
    let z = read zio i in
    (x, y, z)
  end,
  begin fun (x,y,z) o ->
    write xio x o;
    write yio y o;
    write zio z o
  end
;;
(* ***)
(*** io_quadruple *)
let io_quadruple : 'a literate -> 'b literate -> 'c literate -> 'd literate -> ('a * 'b * 'c * 'd) literate =
  fun
    ((xr,xw) as xio)
    ((yr,yw) as yio)
    ((zr,zw) as zio)
    ((tr,tw) as tio)
  ->
  begin fun i ->
    let x = read xio i in
    let y = read yio i in
    let z = read zio i in
    let t = read tio i in
    (x, y, z, t)
  end,
  begin fun (x,y,z,t) o ->
    write xio x o;
    write yio y o;
    write zio z o;
    write tio t o
  end
;;
(* ***)
(*** io_collection *)
let io_collection : ('thing,'bag) collection_io_spec -> 'bag literate =
  fun (xio, iterator, builder) ->
  begin fun i ->
    read_this_tag i T_Collection;
    let (add, get) = builder () in
    let rec loop () =
      match read_tag i with
      | T_End -> get ()
      | T_More ->
          let x = read xio i in
          add x;
          loop ()
      | _ -> raise Unknown_tag
    in
    loop ()
  end,
  begin fun xb o ->
    write_tag o T_Collection;
    iterator xb
      begin fun x ->
        write_tag o T_More;
        write xio x o
      end;
    write_tag o T_End
  end
;;
(* ***)
(*** io_list *)
let io_list : 'a literate -> 'a list literate = fun ((xr,xw) as xio) ->
  begin fun i ->
    read_this_tag i T_List;
    let rec loop () =
      match read_tag i with
      | T_End -> []
      | T_More ->
          let x = read xio i in
          x :: loop ()
      | tag -> raise (Error(sf "Expecting <more> or <end>, got <%s>" (string_of_tag tag)))
    in
    loop ()
  end,
  begin fun xl o ->
    write_tag o T_List;
    let rec loop = function
      | [] -> write_tag o T_End
      | x :: rest ->
          write_tag o T_More;
          write xio x o;
          loop rest
    in
    loop xl
  end
;;
(* ***)
(*** io_array *)
let io_array : 'a literate -> 'a array literate = fun ((xr,xw) as xio) ->
  begin fun i ->
    read_this_tag i T_Array;
    let m = read_int i in
    let x =
      Array.init m
        begin fun _ ->
          read xio i
        end
    in
    read_this_tag i T_End;
    x
  end,
  begin fun x o ->
    write_tag o T_Array;
    write_int o (Array.length x);
    Array.iter
      begin fun x ->
        write xio x o
      end
      x;
    write_tag o T_End
  end
;;
(* ***)
(*** io_hashtbl *)
let io_hashtbl : 'a literate -> 'b literate -> ('a,'b) Hashtbl.t literate = fun ((xr,xw) as xio) ((yr,yw) as yio) ->
  begin fun i ->
    read_this_tag i T_Hashtbl;
    let m = read_int i in
    let h = Hashtbl.create m in
    for j = 1 to m do
      let x = read xio i in
      let y = read yio i in
      Hashtbl.add h x y
    done;
    read_this_tag i T_End;
    h
  end,
  begin fun x o ->
    write_tag o T_Hashtbl;
    write_int o (Hashtbl.length x);
    Hashtbl.iter
      begin fun x y ->
        write xio x o;
        write yio y o
      end
      x;
    write_tag o T_End
  end
;;
(* ***)
(*** io_record *)
let io_record : 'a record_io_spec -> 'a -> 'a literate =
  fun rl initial ->
    begin fun i ->
      read_this_tag i T_Record;
      let rec loop1 current =
        match read_token i with
        | Ident field ->
          let rec loop2 = function
            | [] ->
                loss i;
                loop1 current
            | (field',input,_)::rest ->
                if field = field' then
                  loop1 (input current i)
                else
                  loop2 rest
          in
          loop2 rl
        | Tag T_End -> current
        | t -> raise (Error(sf "Expecting ident or tag in record, not %S" (string_of_token t)))
      in
      let r = loop1 initial in
      r
    end,
    begin fun x o ->
      write_tag o T_Record;
      List.iter
        begin fun (field,_,output) ->
          write_ident o field;
          output x o
        end
        rl;
      write_tag o T_End
    end
;;
(* ***)
(*** io_sum *)
let io_sum : 'a sum_io_spec -> 'a literate =
  fun (sr,sw) ->
  begin fun i ->
    read_this_tag i T_Sum;
    let tag = read_ident i in
    try
      sr tag i
    with
    | Unknown_tag -> raise (Error(sf "Tag %S unknown" tag))
  end,
  begin fun x o ->
    let (tag, f) = sw x in
    write_tag o T_Sum;
    write_ident o tag;
    f o
  end
;;
(* ***)
(*** io_option *)
let io_option : 'a literate -> 'a option literate =
  fun ((xr,xw) as xio) ->
    io_sum
      begin
        begin function
          | "none" -> fun i -> None
          | "some" -> fun i -> Some(read xio i)
          | _      -> raise Unknown_tag
        end,
        begin function
          | None   -> "none", write io_unit ()
          | Some x -> "some", write xio x
        end
      end
;;
(* ***)
