(* Cmd_query *)
(* The EDOS project                                                 *)
(* Copyright (C)2000-2006 Berke Durak                               *)
(* Released under the GNU Lesser General Public License version 2.1 *)

open Unix
open Log
open Html
open Http
open Debug
open Db
open Command
open Sessions
open Globals
open Body_properties
open Xml
module C = Configuration
;;

(*** make_parse_error_paragraph *)
let make_parse_error_paragraph w i j x =
  let m = String.length w in
  P begin
    if i = j then
      if i >= m - 1 then
        Seq[T "There is a parse error at the end of your query.";
            Div("query",[
              P(Seq[T w;
                  Span("highlight", T "?")])])]
      else
        Seq[T (sf "There is a parse error at character %d in your query: %s."
                  (i + 1) x);
          Div("query",[
            P(Seq[
              if i > 0 then T(String.sub w 0 (min (m - 1) i)) else Nop;
              if i < m then Span("highlight", T(String.sub w i 1)) else Nop;
              if i + 1 < m then T(String.sub w (j + 1) (m - j - 1)) else Nop])])]
    else
      begin
        Seq[T (sf "There is a parse error between characters %d and %d in your query: %s."
                  (i + 1) (j + 1) x);
            Div("query",
              [P(Seq[
                if i > 0 then T(String.sub w 0 (min (m - 1) i)) else Nop;
                if i < m then
                  Span("highlight",
                       T(String.sub w i ((min (m - 1) j) - i + 1)))
                else
                  Nop;
                if j + 1 < m then
                  T(String.sub w (j + 1) (m - j - 1))
                else
                  Nop])])]
     end
  end
;;
(* ***)
(*** under_lock *)
let under_lock s_info f =
  Mutex.lock s_info.Sessions.si_mutex;
  Util.wrap
    s_info.Sessions.si_mutex
    Mutex.unlock
    begin fun _ -> f () end
;;
(* ***)
(*** bundle *)
let bundle : bundle_generator = fun db config ->
  (*let global = !Globals.global in
    let sessions = global#sessions in*)
  object(self)
    (*** intializer *)
    initializer ()
    (* ***)
    (*** dql *)
    method private dql ~session:(user, s_info) form =
      let dql = Form.get form ~default:"" "q" in
      under_lock s_info
        begin fun () ->
          let result =
            if dql = "" then
              ""
            else
              try
                s_info.si_bf#clear;
                let query = Processor.query_of_string dql in
                s_info.si_processor#do_query query;
                let txt = s_info.si_bf#get in
                txt
              with
              | Processor.Parse_error(i,j,x) ->
                  let b = Buffer.create 256 in
                  Error.show_parse_error ~out:(Buffer.add_string b) i j x dql;
                  Buffer.contents b
              | x -> sf "Exception: %s" (Printexc.to_string x)
          in
          let x = N("response",[],[ N("output", [], [A(S result)])]) in
          Http.Okay(Http.Text_Xml, Xml x)
        end
    (* ***)
    (*** query *)
    method private query ~session:(user, s_info) form =
      let dql = Form.get form ~default:"" "q" in
      let title = "Query" in
      under_lock s_info
        begin fun () ->
          let result =
            if dql = "" then
              Nop
            else
              try
                s_info.si_bf#clear;
                let query = Processor.query_of_string dql in
                s_info.si_processor#do_query query;
                let txt = s_info.si_bf#get in
                T txt
              with
              | Processor.Parse_error(i,j,x) -> Div("results error",[make_parse_error_paragraph dql i j x])
              | x -> T(sf "Exception: %s" (Printexc.to_string x))
          in
          let doc =
            P(Seq[
                T "Enter your DQL query:";
                With(["id","zonk"],
                  Div("gogol",[
                    Form(GET, "query",
                      Seq[
                        T "DQL expression: ";
                        I_text     ("q", dql, Some 80, Some 512);
                        I_button   ("e", "Evaluate");
                        I_hidden   ("sid", s_info.Sessions.si_sid);
                      ])
                  ]));
                P result])
          in
          raise (Body_properties(title,doc,[(* "onload", "alert('shit')" *)]))
        end
    (* ***)
    (*** console *)
    method private console ~session:(user, s_info) form =
      let title = "Console" in
      let c = "default" in
      let div ?(extra=[]) id (l : html_element list) = With(("id", id) :: extra, Div(c, l)) in
      let doc =
        Seq[
          Script("text/javascript", "/javascript/console.js");
          div "header"
            [ div ~extra:["sid", s_info.si_sid] "console" [
              div "messages" [];
              div "session" [
                div "prompt" [
                  With(["onsubmit", "return CLI__submitted()"],
                    Form(GET, "nothing",
                      Seq[
                        TT "> ";
                        With(["id", "command"],
                             I_text     ("command", "", Some 80, Some 512));
                      ])
                  )
                ]
              ]
            ];
            div "status" [
              With(["id","indicator"], P(T "Welcome..."));
            ];
            div "table" [
              div "tabs" [];
              div "contents" []
            ]
           ]
        ]
      in
      raise (Body_properties(title,doc,["onload", "start()"]))
    (* ***)
    method shutdown = ()
    method table = [
      "query",      Html(Logged_in self#query);
      "dql",        Raw(Logged_in self#dql);
      "console",    Html(Logged_in self#console)
    ]
end
;;
(* ***)
